/*
 * Copyright 2008 The Closure Compiler Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.javascript.jscomp;

import com.google.common.collect.Maps;
import static com.google.javascript.jscomp.JsMessage.Style.RELAX;
import static com.google.javascript.jscomp.JsMessageVisitor.MESSAGE_TREE_MALFORMED;
import static com.google.javascript.jscomp.ReplaceMessages.BUNDLE_DOES_NOT_HAVE_THE_MESSAGE;
import com.google.javascript.rhino.Node;

import junit.framework.TestCase;

import java.util.Locale;
import java.util.Map;

/**
 * Test which checks that replacer works correctly.
 *
 * @author anatol@google.com (Anatol Pomazau)
 */
public class ReplaceMessagesTest extends TestCase {

  private Map<String, JsMessage> messages;
  private Compiler compiler;
  private boolean strictReplacement;

  @Override
  protected void setUp()  {
    messages = Maps.newHashMap();
    strictReplacement = false;
  }

  public void testReplaceSimpleMessage() {
    registerMessage(new JsMessage.Builder("MSG_A")
        .appendStringPart("Hi\nthere")
        .build());

    assertOutputEquals("var MSG_A = goog.getMsg('asdf');",
        "var MSG_A=\"Hi\\nthere\"");
  }

  public void testNameReplacement()  {
    registerMessage(new JsMessage.Builder("MSG_B")
        .appendStringPart("One ")
        .appendPlaceholderReference("measly")
        .appendStringPart(" ph")
        .build());

    assertOutputEquals(
        "var MSG_B = goog.getMsg('asdf {$measly}', {measly: x});",
        "var MSG_B=\"One \"+(x+\" ph\")");
  }

  public void testGetPropReplacement()  {
    registerMessage(new JsMessage.Builder("MSG_C")
        .appendPlaceholderReference("amount")
        .build());

    assertOutputEquals(
        "var MSG_C = goog.getMsg('${$amount}', {amount: a.b.amount});",
        "var MSG_C=a.b.amount");
  }

  public void testFunctionCallReplacement()  {
    registerMessage(new JsMessage.Builder("MSG_D")
        .appendPlaceholderReference("amount")
        .build());

    assertOutputEquals(
        "var MSG_D = goog.getMsg('${$amount}', {amount: getAmt()});",
        "var MSG_D=getAmt()");
  }

  public void testMethodCallReplacement()  {
    registerMessage(new JsMessage.Builder("MSG_E")
        .appendPlaceholderReference("amount")
        .build());

    assertOutputEquals(
        "var MSG_E = goog.getMsg('${$amount}', {amount: obj.getAmt()});",
        "var MSG_E=obj.getAmt()");
  }

  public void testHookReplacement()  {
    registerMessage(new JsMessage.Builder("MSG_F")
        .appendStringPart("#")
        .appendPlaceholderReference("amount")
        .appendStringPart(".")
        .build());

    assertOutputEquals(
        "var MSG_F = goog.getMsg('${$amount}', {amount: (a ? b : c)});",
        "var MSG_F=\"#\"+((a?b:c)+\".\")");
  }

  public void testAddReplacement()  {
    registerMessage(new JsMessage.Builder("MSG_G")
        .appendPlaceholderReference("amount")
        .build());

    assertOutputEquals(
        "var MSG_G = goog.getMsg('${$amount}', {amount: x + ''});",
        "var MSG_G=x+\"\"");
  }

  public void testPlaceholderValueReferencedTwice()  {
    registerMessage(new JsMessage.Builder("MSG_H")
        .appendPlaceholderReference("dick")
        .appendStringPart(", ")
        .appendPlaceholderReference("dick")
        .appendStringPart(" and ")
        .appendPlaceholderReference("jane")
        .build());

    assertOutputEquals(
        "var MSG_H = goog.getMsg('{$dick}{$jane}', {jane: x, dick: y});",
        "var MSG_H=y+(\", \"+(y+(\" and \"+x)))");
  }

  public void testPlaceholderNameInLowerCamelCase()  {
    registerMessage(new JsMessage.Builder("MSG_I")
        .appendStringPart("Sum: $")
        .appendPlaceholderReference("amtEarned")
        .build());

    assertOutputEquals(
        "var MSG_I = goog.getMsg('${$amtEarned}', {amtEarned: x});",
        "var MSG_I=\"Sum: $\"+x");
  }

  public void testQualifiedMessageName()  {
    registerMessage(new JsMessage.Builder("MSG_J")
        .appendStringPart("One ")
        .appendPlaceholderReference("measly")
        .appendStringPart(" ph")
        .build());

    assertOutputEquals(
        "a.b.c.MSG_J = goog.getMsg('asdf {$measly}', {measly: x});",
        "a.b.c.MSG_J=\"One \"+(x+\" ph\")");
  }

  public void testSimpleMessageReplacementMissing()  {
    assertOutputEquals("var MSG_E = 'd*6a0@z>t';", "var MSG_E=\"d*6a0@z>t\"");
  }

  public void testStrictModeAndMessageReplacementAbsentInBundle()  {
    strictReplacement = true;
    process("var MSG_E = 'Hello';");
    assertEquals(1, compiler.getErrors().length);
    assertEquals(BUNDLE_DOES_NOT_HAVE_THE_MESSAGE,
        compiler.getErrors()[0].getType());
  }

  public void testStrictModeAndMessageReplacementAbsentInNonEmptyBundle()  {
    registerMessage(new JsMessage.Builder("MSG_J")
        .appendStringPart("One ")
        .appendPlaceholderReference("measly")
        .appendStringPart(" ph")
        .build());

    strictReplacement = true;
    process("var MSG_E = 'Hello';");
    assertEquals(1, compiler.getErrors().length);
    assertEquals(BUNDLE_DOES_NOT_HAVE_THE_MESSAGE,
        compiler.getErrors()[0].getType());
  }

  public void testFunctionReplacementMissing()  {
    assertOutputEquals("var MSG_F = function() {return 'asdf'};",
        "var MSG_F=function(){return\"asdf\"}");
  }

  public void testFunctionWithParamReplacementMissing()  {
    assertOutputEquals(
        "var MSG_G = function(measly) {return 'asdf' + measly};",
        "var MSG_G=function(measly){return\"asdf\"+measly}");
  }

  public void testPlaceholderNameInLowerUnderscoreCase()  {
    process("var MSG_J = goog.getMsg('${$amt_earned}', {amt_earned: x});");

    assertEquals(1, compiler.getErrors().length);
    JSError error = compiler.getErrors()[0];
    assertEquals(MESSAGE_TREE_MALFORMED, error.getType());
  }

  public void testBadPlaceholderReferenceInReplacement()  {
    registerMessage(new JsMessage.Builder("MSG_K")
        .appendPlaceholderReference("amount")
        .build());

    process("var MSG_K = goog.getMsg('Hi {$jane}', {jane: x});");

    assertEquals(1, compiler.getErrors().length);
    JSError error = compiler.getErrors()[0];
    assertEquals(MESSAGE_TREE_MALFORMED, error.getType());
  }


  public void testLegacyStyleNoPlaceholdersVarSyntax()  {
    registerMessage(new JsMessage.Builder("MSG_A")
        .appendStringPart("Hi\nthere")
        .build());
    assertOutputEquals("var MSG_A = 'd*6a0@z>t';",
        "var MSG_A=\"Hi\\nthere\"");
  }

  public void testLegacyStyleNoPlaceholdersFunctionSyntax()  {
    registerMessage(new JsMessage.Builder("MSG_B")
        .appendStringPart("Hi\nthere")
        .build());

    assertOutputEquals("var MSG_B = function() {return 'asdf'};",
        "var MSG_B=function(){return\"Hi\\nthere\"}");
  }

  public void testLegacyStyleOnePlaceholder()  {
    registerMessage(new JsMessage.Builder("MSG_C")
        .appendStringPart("One ")
        .appendPlaceholderReference("measly")
        .appendStringPart(" ph")
        .build());
    assertOutputEquals(
        "var MSG_C = function(measly) {return 'asdf' + measly};",
        "var MSG_C=function(measly){return\"One \"+(measly+\" ph\")}");
  }

  public void testLegacyStyleTwoPlaceholders()  {
    registerMessage(new JsMessage.Builder("MSG_D")
        .appendPlaceholderReference("dick")
        .appendStringPart(" and ")
        .appendPlaceholderReference("jane")
        .build());
    assertOutputEquals(
        "var MSG_D = function(jane, dick) {return jane + dick};",
        "var MSG_D=function(jane,dick){return dick+(\" and \"+jane)}");
  }

  public void testLegacyStylePlaceholderNameInLowerCamelCase() {
    registerMessage(new JsMessage.Builder("MSG_E")
        .appendStringPart("Sum: $")
        .appendPlaceholderReference("amtEarned")
        .build());
    assertOutputEquals(
        "var MSG_E = function(amtEarned) {return amtEarned + 'x'};",
        "var MSG_E=function(amtEarned){return\"Sum: $\"+amtEarned}");
  }

  public void testLegacyStylePlaceholderNameInLowerUnderscoreCase() {
    registerMessage(new JsMessage.Builder("MSG_F")
        .appendStringPart("Sum: $")
        .appendPlaceholderReference("amt_earned")
        .build());

    // Placeholder named in lower-underscore case (discouraged nowadays)
    assertOutputEquals(
        "var MSG_F = function(amt_earned) {return amt_earned + 'x'};",
        "var MSG_F=function(amt_earned){return\"Sum: $\"+amt_earned}");
  }

  public void testLegacyStyleBadPlaceholderReferenceInReplacemen() {
    registerMessage(new JsMessage.Builder("MSG_B")
        .appendStringPart("Ola, ")
        .appendPlaceholderReference("chimp")
        .build());

    process("var MSG_B = function(chump) {return chump + 'x'};");
    assertEquals(1, compiler.getErrors().length);
    JSError error = compiler.getErrors()[0];
    assertEquals("Message parse tree malformed. "
        + "Unrecognized message placeholder referenced: chimp",
        error.description);
  }

  private void assertOutputEquals(String input, String output) {
    String output1 = process(input);
    JSError[] errors = compiler.getErrors();
    if (errors.length > 0) {
      fail(errors[0].description);
    }

    assertEquals(output, output1);
  }


  private String process(String input) {
    compiler = new Compiler();
    Node root = compiler.parseTestCode(input);
    JsMessageVisitor visitor = new ReplaceMessages(compiler,
        new SimpleMessageBundle(), false, RELAX, strictReplacement);
    visitor.process(null, root);

    return compiler.toSource(root);
  }


  private void registerMessage(JsMessage message) {
    messages.put(message.getKey(), message);
  }

  private class SimpleMessageBundle implements MessageBundle {

    @Override
    public JsMessage getMessage(String id) {
      return messages.get(id);
    }

    @Override
    public Iterable<JsMessage> getAllMessages() {
      throw new UnsupportedOperationException();
    }

    @Override
    public JsMessage.IdGenerator idGenerator() {
      return null;
    }

    @Override
    public Locale getLocale() {
      return Locale.getDefault();
    }
  }
}
